| package |
package := Package name: 'GlorpUnitOfWork'.
package paxVersion: 1;
	basicComment: 'Ths package defines the Glorp UnitOfWork for an in-memory transaction mechanism. Objects are modified inside a unit of work, rather than written individually, and when the unit of work concludes, all are written to the database at once.

This is part of the GLORP system, an open-source library (LGPL(S)) for persisting Smalltalk objects in relational databases. For more information see the #notice property, the accompanying files and/or the web sites http://www.glorp.org or http://www.sourceforge.net/projects/glorp 
'.

package basicPackageVersion: 'image'.


package classNames
	add: #FieldValueWrapper;
	add: #GlorpRowSorter;
	add: #GlorpUnitOfWorkPackage;
	add: #MultipleRowMapKey;
	add: #ObjectTransaction;
	add: #RowMap;
	add: #RowMapForMementos;
	add: #RowMapKey;
	add: #TableSorter;
	add: #TransformingFieldValueWrapper;
	add: #UnitOfWork;
	yourself.

package methodNames
	add: #DatabaseRow -> #at:forcePut:;
	add: #DatabaseRow -> #at:put:;
	add: #DatabaseRow -> #numberOfFields;
	add: #DatabaseRow -> #unifyField:withField:inRow:;
	add: #Join -> #condenseBooleanEquality;
	add: #Join -> #condenseBooleanEqualityExpressions;
	add: #Join -> #isConstant:;
	add: #Join -> #mapFromSource:andTarget:intoRowsIn:;
	add: #Join -> #rewriteEquality;
	yourself.

package binaryGlobalNames: (Set new
	yourself).

package globalAliases: (Set new
	yourself).

package setPrerequisites: (IdentitySet new
	add: '..\..\Object Arts\Dolphin\Base\Dolphin';
	add: 'GlorpDatabase';
	add: 'GlorpExpressions';
	add: 'GlorpMisc';
	yourself).

package!

"Class Definitions"!

Object subclass: #FieldValueWrapper
	instanceVariableNames: 'contents hasValue containedBy'
	classVariableNames: ''
	poolDictionaries: ''
	classInstanceVariableNames: ''!
Object subclass: #GlorpRowSorter
	instanceVariableNames: 'childSelector queue sorted current items childCache'
	classVariableNames: ''
	poolDictionaries: ''
	classInstanceVariableNames: ''!
Object subclass: #GlorpUnitOfWorkPackage
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	classInstanceVariableNames: ''!
Object subclass: #MultipleRowMapKey
	instanceVariableNames: 'keys'
	classVariableNames: ''
	poolDictionaries: ''
	classInstanceVariableNames: ''!
Object subclass: #ObjectTransaction
	instanceVariableNames: 'undoMap collectionMementos'
	classVariableNames: ''
	poolDictionaries: ''
	classInstanceVariableNames: ''!
Object subclass: #RowMap
	instanceVariableNames: 'rowDictionary hasDeletions'
	classVariableNames: ''
	poolDictionaries: ''
	classInstanceVariableNames: ''!
Object subclass: #RowMapKey
	instanceVariableNames: 'key1 key2'
	classVariableNames: ''
	poolDictionaries: ''
	classInstanceVariableNames: ''!
Object subclass: #TableSorter
	instanceVariableNames: 'orderedTables tables visitedTables'
	classVariableNames: ''
	poolDictionaries: ''
	classInstanceVariableNames: ''!
Object subclass: #UnitOfWork
	instanceVariableNames: 'session transaction deletedObjects newObjects rowMap commitPlan deletePlan rowsByTable succeeded linkRowsForUpdate commitPhase deleting'
	classVariableNames: ''
	poolDictionaries: ''
	classInstanceVariableNames: ''!
FieldValueWrapper subclass: #TransformingFieldValueWrapper
	instanceVariableNames: 'forwardTransformation inverseTransformation innerWrapper'
	classVariableNames: ''
	poolDictionaries: ''
	classInstanceVariableNames: ''!
RowMap subclass: #RowMapForMementos
	instanceVariableNames: 'correspondenceMap reversedCorrespondenceMap collectionMementos'
	classVariableNames: ''
	poolDictionaries: ''
	classInstanceVariableNames: ''!

"Global Aliases"!


"Loose Methods"!

!DatabaseRow methodsFor!

at: aField forcePut: aValue 
	"Write the value. If there's a previous value, overwrite it. Use with extreme caution."
	| wrapper |
	aValue isGlorpExpression ifTrue: [self error: 'cannot store expressions in rows'].
	aField table == self table ifFalse: [self error: 'Invalid table'].
	wrapper := self contentsAt: aField ifAbsentPut: [FieldValueWrapper new].
	wrapper privateSetContents: aValue.
	^wrapper.!

at: aField put: aValue 
	
	"For generated fields, we expect the real value to be provided later by the database, so don't write a nil value"
	
	| wrapper |
"	aValue isGlorpExpression ifTrue: [self error: 'cannot store expressions in rows']."
	aField table == self table ifFalse: [self error: 'Invalid table'].
	wrapper := self contentsAt: aField ifAbsentPut: [FieldValueWrapper new].
	(aValue isNil and: [aField isGenerated]) ifFalse: [wrapper contents: aValue].
	wrapper isNowContainedBy: self and: aField.
	^wrapper.!

numberOfFields
	^contents
		inject: 0
		into:
			[:sum :each | sum + (each == self class emptySlot ifTrue: [0] ifFalse: [1])].!

unifyField: eachSourceField withField: eachTargetField inRow: targetRow
	| sourceWrapper targetWrapper |
	sourceWrapper := self wrapperAt: eachSourceField ifAbsent: [nil].
	targetWrapper := targetRow wrapperAt: eachTargetField ifAbsent: [nil].
	sourceWrapper isNil
		ifTrue:
			[targetWrapper isNil
				ifTrue:
					[targetWrapper := FieldValueWrapper new.
					targetRow wrapperAt: eachTargetField put: targetWrapper].
			^self wrapperAt: eachSourceField put: targetWrapper].
	targetWrapper isNil ifTrue: [^targetRow wrapperAt: eachTargetField put: sourceWrapper].
	sourceWrapper unifyWith: targetWrapper.! !
!DatabaseRow categoriesFor: #at:forcePut:!private! !
!DatabaseRow categoriesFor: #at:put:!accessing!private! !
!DatabaseRow categoriesFor: #numberOfFields!accessing!private! !
!DatabaseRow categoriesFor: #unifyField:withField:inRow:!api!private! !

!Join methodsFor!

condenseBooleanEquality
	^self.!

condenseBooleanEqualityExpressions
	^self.!

isConstant: aTarget
	"The target can be either a constant (which gets turned into a ConstantExpression) or (usually) a DatabaseField, representing a parameter to the query"

	aTarget isDatabaseField ifTrue: [^false].
	(aTarget isGlorpExpression and: [aTarget isConstantExpression not]) ifTrue: [^false].
	^true!

mapFromSource: sourceObject andTarget: targetObject intoRowsIn: aRowMap
	(sourceObject isNil or: [targetObject isNil]) ifTrue: [^self].
	1
		to: sources size
		do:
			[:i | 
			| eachSourceField eachTargetField |
			eachSourceField := sources at: i.
			eachTargetField := targets at: i.
			(self isConstant: eachSourceField)
				ifTrue: [aRowMap at: eachTargetField key: targetObject put: eachSourceField value]
				ifFalse:
					[aRowMap
						unifyField: eachSourceField
						key: sourceObject
						withField: eachTargetField
						key: targetObject]].!

rewriteEquality
	^self.! !
!Join categoriesFor: #condenseBooleanEquality!preparing!private! !
!Join categoriesFor: #condenseBooleanEqualityExpressions!preparing!private! !
!Join categoriesFor: #isConstant:!converting!public! !
!Join categoriesFor: #mapFromSource:andTarget:intoRowsIn:!api!private! !
!Join categoriesFor: #rewriteEquality!preparing!private! !

"End of package definition"!

"Source Globals"!

"Classes"!

FieldValueWrapper guid: (GUID fromString: '{5CB380DE-C1C2-4DD8-BC98-A32ADAB336B1}')!
FieldValueWrapper comment: 'When we are creating our rowmap, we need to be able to unify values. That is, we express the constraint that field A in table T1 must be equal to field B in table T2. And we have to be able to do this whether or not we have yet assigned a value to either of these fields, and whether or not other equality constraints have already been imposed on one or the other of these fields. We do this by holding the values in a wrapper. When we unify them, we replace one wrapper with the other, so that both rows now refer to the same value. In order to efficiently find where things are contained, we hold a collection of the containing rows.

Instance Variables:
	containedBy	<(Association key: DatabaseField value: (IdentitySet of: DatabaseRow)) | (IdentityDictionary of: Same)>	The rows that contain this wrapper. Using a trick similar to #dependents, we optimize the containedByCollection to be an association if we are only contained in one row (the most common case). Otherwise we have a dictionary. The containing rows are indexed by field, because when we''re updating the row to replace the wrapper, we need to know which field to replace. Also, a single row could conceivably contain the same wrapper in two different fields, and both need to be updated.
	contents	<Object>	The value we contain
	hasValue	<Boolean>	Has our value been set or not. We can''t just test for nil in contents, because nil is a valid value.

'!
!FieldValueWrapper categoriesForClass!Unclassified! !
!FieldValueWrapper methodsFor!

allInnerWrappers
	^#().!

basicContents
	"Return the un-transformed contents"
	^self contents.!

containedBy

	^containedBy.!

containsWrapper: aFieldValueWrapper
	^self == aFieldValueWrapper.!

contents

	^contents.!

contents: anObject

	(hasValue and: [contents ~= anObject]) ifTrue: [self error: 'Inconsistent values in field'].
	self privateSetContents: anObject.!

hasValue

	^hasValue.!

initialize

	hasValue := false.!

innerWrapper
	^self.!

isNowContainedBy: aRow and: aField
	"Optimize this for memory consumption with the common case of one container. If we're contained by nothing, make it a single association. Otherwise, use a dictionary"
	| shortCircuit thisFieldsEntries |
	shortCircuit := false.
	containedBy == nil ifTrue: [containedBy := aField -> ((IdentitySet new: 2) add: aRow; yourself). ^self].
	containedBy class == Association 
		ifTrue: [
		containedBy key == aField
			ifTrue: [thisFieldsEntries := containedBy value]
			ifFalse: [ | newContainedBy |
				newContainedBy := IdentityDictionary new. 
				newContainedBy at: containedBy key put: containedBy value.
				containedBy := newContainedBy]].
	thisFieldsEntries isNil ifTrue: [
		thisFieldsEntries := containedBy at: aField ifAbsentPut: [shortCircuit := true. (IdentitySet new: 2) add: aRow; yourself]].
	shortCircuit ifTrue: [^self].
	(thisFieldsEntries includes: aRow) ifFalse: [thisFieldsEntries add: aRow].!

printOn: aStream

	aStream 
		nextPutAll: '<<'.
	self hasValue ifTrue: [aStream print: self contents].
	aStream
		nextPutAll: '>>'.!

privateSetContents: anObject
	contents := anObject.
	hasValue := true.!

replaceWith: aFieldValueWrapper
	containedBy isNil ifTrue: [^nil].
	containedBy
		keysAndValuesDo:
			[:eachField :eachListOfRows | eachListOfRows do: [:eachRow | eachRow wrapperAt: eachField put: aFieldValueWrapper]].!

unifyWith: targetWrapper
	(self containsWrapper: targetWrapper) ifTrue: [^self].
	(targetWrapper containsWrapper: self) ifTrue: [^self].
	targetWrapper hasValue ifFalse: [^targetWrapper replaceWith: self].
	self hasValue ifFalse: [^self replaceWith: targetWrapper].
	self contents = targetWrapper contents
		ifFalse: [self error: 'Conflicting values in rows'].
	targetWrapper replaceWith: self.! !
!FieldValueWrapper categoriesFor: #allInnerWrappers!accessing!private! !
!FieldValueWrapper categoriesFor: #basicContents!accessing!private! !
!FieldValueWrapper categoriesFor: #containedBy!public! !
!FieldValueWrapper categoriesFor: #containsWrapper:!private!unifying! !
!FieldValueWrapper categoriesFor: #contents!public! !
!FieldValueWrapper categoriesFor: #contents:!public! !
!FieldValueWrapper categoriesFor: #hasValue!public! !
!FieldValueWrapper categoriesFor: #initialize!public! !
!FieldValueWrapper categoriesFor: #innerWrapper!accessing!private! !
!FieldValueWrapper categoriesFor: #isNowContainedBy:and:!containing!private! !
!FieldValueWrapper categoriesFor: #printOn:!public! !
!FieldValueWrapper categoriesFor: #privateSetContents:!public! !
!FieldValueWrapper categoriesFor: #replaceWith:!private!unifying! !
!FieldValueWrapper categoriesFor: #unifyWith:!private!unifying! !

!FieldValueWrapper class methodsFor!

new

	^super new initialize.! !
!FieldValueWrapper class categoriesFor: #new!public! !

GlorpRowSorter guid: (GUID fromString: '{48FE1017-F41A-4DEE-894E-3915880A34E3}')!
GlorpRowSorter comment: 'This sorts a set of rows to determine the order in which they should be inserted into the database. It does this using a breadth first topological sort. What that means is that instead of traversing the graph depth first, as is normal for topological, we traverse it breadth first. This gives us the desirable property of keeping the child rows together as much as possible. Since the initial input will already have been sorted by the static table ordering that is implied by the foreign key constraints, this will also keep rows from the same table together, unless the sort algorithm finds a reason the order needs to be changed. This maximizes the opportunities for using optimizations like Oracle array binding, which want to write groups of similar rows all at once. Note that we can''t just use the table order directly, because there may be row-level issues. That is, rows may need to be written before other rows in the same table, or even if there is a table constraint, it may only be relevant for a small number of rows.

Instance Variables:
	childCache	<IdentityDictionary from: Object to: Collection>	The calculation of children may be expensive, and may have to be done more than once for an item, so we cache the information here.
	childSelector	<Symbol | BlockClosure>	The zero-argument selector which we perform on an item, or the one-argument block which we evaluate in order to get the children.
	current	<DatabaseRow | GlorpBreadthFirstTopologicalSortItem>	The item we are currently examining.
	items	<ReadStream>	A stream on the items we are intended to sort.
	queue	<OrderedCollection>	This maintains the breadth first order. So, when we encounter an item, we add its children to the end of the queue. The Sorter processes the items in order off the queue, resulting in a breadth first traversal.
	sorted	<OrderedCollection>	The resulting collection, in sorted order.

'!
!GlorpRowSorter categoriesForClass!Unclassified! !
!GlorpRowSorter methodsFor!

childrenOf: anItem
	^childCache at: anItem ifAbsentPut: [
		childSelector isSymbol
			ifTrue: [anItem perform: childSelector]
			ifFalse: [childSelector value: anItem]].!

currentChildren
	^self childrenOf: current.!

currentItemIsDone
	^current status = #done.!

currentItemIsInProgress
	^current status = #visitedOnce.!

currentItemIsNew
	^current status = #notVisited.!

getChildrenVia: aSelector
	childSelector := aSelector.!

hasSomethingToProcess
	self next.
	^current notNil.!

initialize
	queue := OrderedCollection new: 100.
	sorted := OrderedCollection new: 100.
	childCache := IdentityDictionary new: 100.!

initialVisit
	| children |
	children := self currentChildren select: [:each  | each status == #notVisited].
	children do: [:each | each parent: current].
	queue addAll: children.
	current status: #visitedOnce.
	children isEmpty ifTrue: [queue add: current].!

next
	queue isEmpty ifTrue: [
		| candidate |
		candidate := items detect: [:each | each status == #notVisited] ifNone: [nil].
		candidate isNil ifTrue: [
			"We're probably at the end, but check again for part-way done nodes, which can happen if there are heavy cycles"
			items reset. 
			candidate := items detect: [:each | each status == #visitedOnce] ifNone: [nil]].
		candidate isNil ifFalse: [queue add: candidate]].
	current := queue isEmpty ifTrue: [nil] ifFalse: [queue removeFirst].!

processCurrent
	self currentItemIsNew ifTrue: [^self initialVisit].
	self currentItemIsInProgress ifTrue: [^self secondVisit].
	self currentItemIsDone ifTrue: [^self].
	self error: 'Unknown item status'.!

secondVisit
	sorted add: current.
	current status: #done.
	current parent isNil
		ifFalse:
			[((self childrenOf: current parent)
				allSatisfy: [:each | each status == #done])
				ifTrue: [queue add: current parent].
			current parent: nil].!

sort: aCollection
	items := aCollection readStream.
	aCollection do: [:each | each status: #notVisited].
	[self hasSomethingToProcess] whileTrue: [self processCurrent].
	^sorted.! !
!GlorpRowSorter categoriesFor: #childrenOf:!private!sorting! !
!GlorpRowSorter categoriesFor: #currentChildren!private!sorting! !
!GlorpRowSorter categoriesFor: #currentItemIsDone!private!sorting! !
!GlorpRowSorter categoriesFor: #currentItemIsInProgress!private!sorting! !
!GlorpRowSorter categoriesFor: #currentItemIsNew!private!sorting! !
!GlorpRowSorter categoriesFor: #getChildrenVia:!accessing!private! !
!GlorpRowSorter categoriesFor: #hasSomethingToProcess!private!sorting! !
!GlorpRowSorter categoriesFor: #initialize!initializing!private! !
!GlorpRowSorter categoriesFor: #initialVisit!private!sorting! !
!GlorpRowSorter categoriesFor: #next!private!sorting! !
!GlorpRowSorter categoriesFor: #processCurrent!private!sorting! !
!GlorpRowSorter categoriesFor: #secondVisit!private!sorting! !
!GlorpRowSorter categoriesFor: #sort:!private!sorting! !

!GlorpRowSorter class methodsFor!

new
	^super new initialize.! !
!GlorpRowSorter class categoriesFor: #new!instance creation!private! !

GlorpUnitOfWorkPackage guid: (GUID fromString: '{1CD0A18B-BB60-4A91-B539-D2737F8B2B4A}')!
GlorpUnitOfWorkPackage comment: 'This class represents package information for the GlorpUnitOfWork package, for systems that don''t have things like package properties to store things like copyright and license information.

LGPL(S)

The code is Copyright (C) 2000-2010  Alan Knight

This is a Smalltalk library, licensed under the LGPL, and distributed WITHOUT
ANY WARRANTY (see below).  However, it is not obvious how some of
the terms and concepts in the license should be sensibly applied to Smalltalk
The various Smalltalk image models often do not directly correspond to 
the ideas of programs and libraries that are linked to form executables. 
We, the authors, would like to clarify our interpretation of the LGPL as it
applies to Smalltalk, and what we permit you to do with this code. We
will sometimes use the term LGPL(S) to distinguish the use of the LGPL 
with these clarifications.

This code is intended to be usable as a library, without the intention to 
restrict the license of the program that uses it. Thus, you may use this library
the same way you would normally use any other Smalltalk library. That is, 
you may load it into an image or otherwise make it available for use in whatever
ways are appropriate to the Smalltalk implementation you are using (including,
but not limited to file-in, linking a shared library, or loading a binary 
representation such as a parcel,  BOSS file, image segment or image component).
You may write code that uses the library, including subclassing library classes
and having the library call back to your code using blocks, the #perform: 
mechanism, or similar mechanisms. You may use this code,  including creating
and distributing packaged images, libraries, development images and executables
that include the library code. We do not consider these activities sufficient
to make the entire program/image/executable or any portion of it beyond the
library itself be considered a derivative work, and in any case we place no
restrictions on what licenses you may use for code that uses this library in
these or similar manners.

If, however, you modify the class definitions or methods of the library, we do
consider those modified sections a derivative work, and as such they are fully subject
to the restrictions described in the LGPL. Basically, you''ll need to release the
source of any modifications.

If you require any further clarification, we''d be happy to provide it. You can contact
the authors c/o Alan Knight, knight@acm.org.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
'!
!GlorpUnitOfWorkPackage categoriesForClass!Unclassified! !
!GlorpUnitOfWorkPackage class methodsFor!

LICENSE
"LGPL(S)

The code is Copyright (C) 2000-2010  Alan Knight

This is a Smalltalk library, licensed under the LGPL, and distributed WITHOUT
ANY WARRANTY (see below).  However, it is not obvious how some of
the terms and concepts in the license should be sensibly applied to Smalltalk
The various Smalltalk image models often do not directly correspond to 
the ideas of programs and libraries that are linked to form executables. 
We, the authors, would like to clarify our interpretation of the LGPL as it
applies to Smalltalk, and what we permit you to do with this code. We
will sometimes use the term LGPL(S) to distinguish the use of the LGPL 
with these clarifications.

This code is intended to be usable as a library, without the intention to 
restrict the license of the program that uses it. Thus, you may use this library
the same way you would normally use any other Smalltalk library. That is, 
you may load it into an image or otherwise make it available for use in whatever
ways are appropriate to the Smalltalk implementation you are using (including,
but not limited to file-in, linking a shared library, or loading a binary 
representation such as a parcel,  BOSS file, image segment or image component).
You may write code that uses the library, including subclassing library classes
and having the library call back to your code using blocks, the #perform: 
mechanism, or similar mechanisms. You may use this code,  including creating
and distributing packaged images, libraries, development images and executables
that include the library code. We do not consider these activities sufficient
to make the entire program/image/executable or any portion of it beyond the
library itself be considered a derivative work, and in any case we place no
restrictions on what licenses you may use for code that uses this library in
these or similar manners.

If, however, you modify the class definitions or methods of the library, we do
consider those modified sections a derivative work, and as such they are fully subject
to the restrictions described in the LGPL. Basically, you'll need to release the
source of any modifications.

If you require any further clarification, we'd be happy to provide it. You can contact
the authors c/o Alan Knight, knight@acm.org.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA"! !
!GlorpUnitOfWorkPackage class categoriesFor: #LICENSE!instance creation!public! !

MultipleRowMapKey guid: (GUID fromString: '{FDFE1247-27CD-4F01-A69E-AFC158C0BBF7}')!
MultipleRowMapKey comment: 'This is a special (and rarely needed) form of row map key that allows an arbitrary number of objects to participate in it.'!
!MultipleRowMapKey categoriesForClass!Unclassified! !
!MultipleRowMapKey methodsFor!

= aRowMapKey

	aRowMapKey class == self class ifFalse: [^false].
	aRowMapKey keys size = keys size ifFalse: [^false].
	"We can't rely on the order being the same because of the obscure but real case of two different related objects with the same identityHash"
	aRowMapKey keys do: [:each |
		(keys includes: each) ifFalse: [^false]].
	^true.!

addAllKeys: aKeyCollection
	keys addAll: aKeyCollection.!

addKey: aKey
	keys add: aKey.!

adjustIn: aCorrespondenceMap
	"Adjust ourselves to refer to the copies rather than the originals"
	| adjusted |
	adjusted := self copy.
	adjusted keys: (keys collect: [:each | aCorrespondenceMap at: each ifAbsent: [each]]).
	^adjusted.!

hash
	"Note: the two things with the same identity hash problem is OK here because if they do then the hashes are (by definition) the same, and the order doesn't matter"

	| increment hash |
	increment := 14 // keys size.
	hash := 0.
	keys keysAndValuesDo: [:index :each | hash := hash bitXor: (each identityHash bitShift: increment * index)].
	^hash!

initialize
	keys := (SortedCollection new: 4) sortBlock: self sortBlock.!

keys
	^keys.!

keys: aCollection
	keys := (aCollection asSortedCollection: self sortBlock).!

printOn: aStream
	aStream nextPutAll: 'KEY('.
	self hash printOn: aStream.
	aStream nextPutAll: '):'.
	keys printOn: aStream.!

reverseAdjustIn: aRowMapForMementos
	"Adjust ourselves to refer to the originals rather than the copies"
	| adjusted |
	adjusted := self copy.
	adjusted keys: (keys collect: [:each | 
		aRowMapForMementos reversedCorrespondenceMap at: each ifAbsent: [each]]).
	^adjusted.!

sortBlock
	^[:a :b | a identityHash <= b identityHash].! !
!MultipleRowMapKey categoriesFor: #=!comparing!private! !
!MultipleRowMapKey categoriesFor: #addAllKeys:!accessing!private! !
!MultipleRowMapKey categoriesFor: #addKey:!accessing!private! !
!MultipleRowMapKey categoriesFor: #adjustIn:!adjusting!private! !
!MultipleRowMapKey categoriesFor: #hash!comparing!public! !
!MultipleRowMapKey categoriesFor: #initialize!initialize-release!private! !
!MultipleRowMapKey categoriesFor: #keys!accessing!private! !
!MultipleRowMapKey categoriesFor: #keys:!accessing!private! !
!MultipleRowMapKey categoriesFor: #printOn:!printing!private! !
!MultipleRowMapKey categoriesFor: #reverseAdjustIn:!adjusting!private! !
!MultipleRowMapKey categoriesFor: #sortBlock!initialize-release!private! !

!MultipleRowMapKey class methodsFor!

new
	"Answer a newly created and initialized instance."

	^super new initialize!

with: key1 with: key2 with: key3
	(key1 isNil or: [key2 isNil or: [key3 isNil]]) ifTrue: [self error: 'Nil key in row map, probably indicates an invalid object structure'].
	^self new 
		addKey: key1;
		addKey: key2;
		addKey: key3;
		yourself.!

withAll: aCollection
	
	^self new 
		addAllKeys: aCollection;
		yourself.! !
!MultipleRowMapKey class categoriesFor: #new!instance creation!private! !
!MultipleRowMapKey class categoriesFor: #with:with:with:!instance creation!private! !
!MultipleRowMapKey class categoriesFor: #withAll:!instance creation!private! !

ObjectTransaction guid: (GUID fromString: '{BA6C883D-AB42-43C3-8EAC-3B59A4692B73}')!
ObjectTransaction comment: 'An ObjectTransaction knows how to remember the state of objects and revert them back to that state later on. It does this by making a *shallow* copy of the registered objects and everything connected to them, and then putting that into an identity dictionary keyed by the originals.

If you have to undo, you push the state from the shallow copies back into the originals.

Yes, that works, and it''s all you have to do. It even handles collections become:ing different sizes.

This is fairly independent of GLORP. You could use this mechanism in general, if you provided your own mechanism for figuring out what to register, or even just uncommented the one in here.

Instance Variables:
	undoMap	<IdentityDictionary>	 The dictionary of originals->copies.

'!
!ObjectTransaction categoriesForClass!Unclassified! !
!ObjectTransaction methodsFor!

abort
	undoMap keysAndValuesDo: [:original :copy | self restoreStateOf: original toThatOf: copy]!

abort: original
	| copy |
	copy := undoMap at: original ifAbsent: [^self].
	self restoreStateOf: original toThatOf: copy!

begin
	self initializeUndoMap!

collectionMementoFor: anObject
	anObject isNil ifTrue: [^#()].
	^collectionMementos at: anObject ifAbsent: [self mementoFor: anObject].!

collectionMementos
	^collectionMementos.!

commit
	self initializeUndoMap!

initialize
	self initializeUndoMap.!

initializeUndoMap

	undoMap := IdentityDictionary new: 1000.
	collectionMementos := IdentityDictionary new: 200.!

isRegistered: anObject 
	"Note: We can never have a situation where a proxy is registered but its contents aren't, so we don't have to worry about that ambiguous case."
	| realObject |
	realObject := self realObjectFor: anObject ifNone: [^false].
	^undoMap includesKey: realObject.!

isShapeOf: original differentThanThatOf: copy

	^original class ~~ copy class or: [original glorpBasicSize ~= copy glorpBasicSize]!

mementoFor: anObject
	^undoMap at: anObject.!

needsCollectionMementoFor: aCollection
	"We can avoid the need for a collection copy if the collection has no internal structure - that is, if we'll be able to loop over a simple shallow copy. This is dialect and collection class dependent"
	^Dialect needsCollectionMementoFor: aCollection.!

realObjectFor: anObject 
	"If this is a proxy, return the contents (if available). Otherwise, return nil"
	^self realObjectFor: anObject ifNone: [nil].!

realObjectFor: anObject ifNone: aBlock
	"If this is a proxy, return the contents (if available). Otherwise, evaluate the block"
	^anObject isGlorpProxy 
		ifTrue: [anObject isInstantiated ifTrue: [anObject getValue] ifFalse: [aBlock value]]
		ifFalse: [anObject]!

register: anObject 
	"Make anObject be a member of the current transaction. Return the object if registered, or nil otherwise"

	| copy realObject |
	(self requiresRegistrationFor: anObject) ifFalse: [^nil].
	realObject := self realObjectFor: anObject ifNone: [^nil].
	copy := self shallowCopyOf: realObject ifNotNeeded: [^nil].
	undoMap at: realObject put: copy.
	self registerTransientInternalsOfCollection: realObject.
	^realObject!

registeredObjectsDo: aBlock
	"Iterate over all our objects. Note that this will include objects without descriptors. Be sure we're iterating over a copy of the keys, because this will add objects to the undoMap. Allow the block to return a collection which we will then add to the list of things to process, until there's nothing further"

	| newAdditions previousVersion |
	newAdditions := undoMap keys.
	previousVersion := newAdditions.
	[newAdditions isEmpty]
		whileFalse:
			[newAdditions do: aBlock.
			newAdditions := OrderedCollection new.
			undoMap
				keysDo:
					[:eachKey | (previousVersion includes: eachKey) ifFalse: [newAdditions add: eachKey]].
			newAdditions isEmpty ifFalse: [previousVersion := undoMap keys]].!

registerTransientInternalsOfCollection: aCollection
	"If this is a collection, then we may need to register any internal structures it has, e.g. an internal array. This is implementation dependent for the collection. We will also explicitly exclude strings"

	aCollection glorpIsCollection ifFalse: [^self].
	aCollection class isBits ifTrue: [^self].
	(self needsCollectionMementoFor: aCollection)
		ifTrue:
			[| copy |
			copy := collectionMementos at: aCollection put: aCollection copy.
			copy class == Dictionary
				ifTrue:
					[1 to: copy basicSize do: [:i | copy basicAt: i put: (copy basicAt: i) copy]	"We need to copy the associations"]].
	aCollection glorpRegisterCollectionInternalsIn: self!

requiresRegistrationFor: anObject

	| realObject |
	realObject := self realObjectFor: anObject ifNone: [^false].
	^(self isRegistered: realObject) not.!

restoreIndexedInstanceVariablesOf: original toThoseOf: copy

	copy glorpStartOfIndexedInstVars to: copy glorpEndOfIndexedInstVars do: [:index |
		original isString 
			ifTrue: [original at: index put: (copy at: index)]
			ifFalse: [original basicAt: index put: (copy basicAt: index)]].!

restoreNamedInstanceVariablesOf: original toThoseOf: copy

	1 to: copy class instSize do: [:index |
		original instVarAt: index put: (copy instVarAt: index)]!

restoreShapeOf: original toThatOf: copy

	| newOriginal |
	(copy class isBits or: [copy class isVariable])
		ifTrue: [newOriginal := copy class basicNew: copy glorpBasicSize]
		ifFalse: [newOriginal := copy class basicNew].
	original become: newOriginal.!

restoreStateOf: original toThatOf: copy 

	(self isShapeOf: original differentThanThatOf: copy) 
		ifTrue: [self restoreShapeOf: original toThatOf: copy].
	self restoreNamedInstanceVariablesOf: original toThoseOf: copy.
	self restoreIndexedInstanceVariablesOf: original toThoseOf: copy!

shallowCopyOf: anObject ifNotNeeded: aBlock

	| copy |
	copy := anObject shallowCopy.
	^copy == anObject 
		ifTrue: [aBlock value]
		ifFalse: [copy]!

undoMap
	^undoMap! !
!ObjectTransaction categoriesFor: #abort!begin/commit/abort!private! !
!ObjectTransaction categoriesFor: #abort:!begin/commit/abort!private! !
!ObjectTransaction categoriesFor: #begin!begin/commit/abort!private! !
!ObjectTransaction categoriesFor: #collectionMementoFor:!accessing!private! !
!ObjectTransaction categoriesFor: #collectionMementos!accessing!private! !
!ObjectTransaction categoriesFor: #commit!begin/commit/abort!private! !
!ObjectTransaction categoriesFor: #initialize!initializing!private! !
!ObjectTransaction categoriesFor: #initializeUndoMap!initializing!private! !
!ObjectTransaction categoriesFor: #isRegistered:!private!registering! !
!ObjectTransaction categoriesFor: #isShapeOf:differentThanThatOf:!private!private/restoring! !
!ObjectTransaction categoriesFor: #mementoFor:!accessing!private! !
!ObjectTransaction categoriesFor: #needsCollectionMementoFor:!private!registering! !
!ObjectTransaction categoriesFor: #realObjectFor:!private!registering! !
!ObjectTransaction categoriesFor: #realObjectFor:ifNone:!private!registering! !
!ObjectTransaction categoriesFor: #register:!private!registering! !
!ObjectTransaction categoriesFor: #registeredObjectsDo:!private!registering! !
!ObjectTransaction categoriesFor: #registerTransientInternalsOfCollection:!public!registering! !
!ObjectTransaction categoriesFor: #requiresRegistrationFor:!private!registering! !
!ObjectTransaction categoriesFor: #restoreIndexedInstanceVariablesOf:toThoseOf:!private!private/restoring! !
!ObjectTransaction categoriesFor: #restoreNamedInstanceVariablesOf:toThoseOf:!private!private/restoring! !
!ObjectTransaction categoriesFor: #restoreShapeOf:toThatOf:!private!private/restoring! !
!ObjectTransaction categoriesFor: #restoreStateOf:toThatOf:!private!private/restoring! !
!ObjectTransaction categoriesFor: #shallowCopyOf:ifNotNeeded:!private!private/registering! !
!ObjectTransaction categoriesFor: #undoMap!accessing!private! !

!ObjectTransaction class methodsFor!

new
	^super new initialize! !
!ObjectTransaction class categoriesFor: #new!instance creation!private! !

RowMap guid: (GUID fromString: '{D8DC8483-F265-4464-95EC-A76D81B64855}')!
RowMap comment: 'This is the central class used for writing. We build up a collection of rows that are to be written, indexed by the objects that "own" them. While any object can contribute values to a row (for example, in a relationship, the value of a field that is a foreign key may come from the primary key of another object) it is a central assumption that each row is owned by a single object. Once we know all the rows, we can compute which ones have changed from their previous database values (see RowMapForMementos) and which fields within them have changed, and then we can write them out.

Instance Variables:
	hasDeletions	<Boolean>	Do we have any deletions. Deletions are marked by a flag in the row, but it''s useful to know quickly if we have any or not.
	rowDictionary	<(Dictionary from: DatabaseTable to: (Dictionary from: Object to: DatabaseRow)>	The rows, indexed by their table, then by the owning object.

'!
!RowMap categoriesForClass!Unclassified! !
!RowMap methodsFor!

additiveDifferencesFrom: aRowMap into: differencesMap
	"Add everything which is in us, but not in aRowMap into differencesMap"
	| newRow |
	self objectsAndRowsDo: [:object :row |
		| correspondingRow |
		correspondingRow := aRowMap
			rowForTable: row table
			withKey: object
			ifAbsent: [DatabaseRow new].
		(row equals: correspondingRow)
			ifFalse:
				[newRow := differencesMap
					addRow: (row withAllFieldsIn: correspondingRow)
					forTable: row table
					withKey: object.
				newRow forDeletion: row forDeletion]].!

addRow: aRow
	^self addRow: aRow forTable: aRow table withKey: aRow owner.!

addRow: aRow forTable: aTable withKey: aKey 
	| submap |
	aKey isGlorpProxy ifTrue: [self halt].
	submap := self subMapForTable: aTable withKey: aKey.
	submap at: aKey put: aRow.
	^aRow.!

adjustForMementos: objects
	^self.!

at: aField key: anObject put: aValue
	(self findOrAddRowForTable: aField table withKey: anObject) at: aField put: aValue.!

collapseMatchingInsertsAndDeletes
	^rowDictionary
		inject: IdentitySet new
		into:
			[:sum :eachTablesRowList | 
			sum addAll: (self collapseMatchingInsertsAndDeletesFor: eachTablesRowList).
			sum].!

collapseMatchingInsertsAndDeletesFor: eachTablesRows
	"Find any inserts and deletes for these rows which have the same primary key. We expect these to be link table rows, because for rows that map to objects, object identity should ensure this never happens. Force the insert to be an update instead (by returning it to the calling method, which will force that to happen) and remove the deletion"
	| deletions updates matchedDeletions deletion nonGeneratedFieldSize |
	(self isSuitableForCollapsing: eachTablesRows) ifFalse: [^#()].
	deletions := Dictionary new: 50.
	updates := OrderedCollection new.
	eachTablesRows do: [:each | each forDeletion ifTrue: [
		deletions at: (each primaryKey) put: each]].
	deletions isEmpty ifTrue: [^#()].
	matchedDeletions := IdentitySet new: deletions size.

	nonGeneratedFieldSize := (eachTablesRows detect: [:each | true]) nonGeneratedFields size.
	eachTablesRows do: [:each |
		| pk |
		pk := each primaryKey.
		((deletions includesKey: pk) and: [each forDeletion not]) ifTrue: [
			each numberOfFields = nonGeneratedFieldSize ifFalse: [self error: 'link table entry requiring update is only partially populated'].
			deletion := deletions at: pk.
			(matchedDeletions includes: deletion)
				ifFalse: [
					"Now here's a very special case. If there are no non-primary key fields, then we collapse this down to a no-op. We've deleted and re-inserted with exactly the same information. However, we still leave it in the list of updated rows, because we want it to be put into the session cache if it's not already there at the end."
					each table primaryKeyFields size = each fields size
						ifTrue: [self privateDeleteRow: each].
					updates add: each.
					matchedDeletions add: deletion.
					self privateDeleteRow: deletion]]].
	^updates!

collectionMementoFor: aCollection

	^aCollection.!

compress
	"Remove all deleted rows and rebuild"
	| newDictionary |
	hasDeletions ifFalse: [^self].
	Transcript cr; show: 'compressing'.
	hasDeletions := false.
	newDictionary := IdentityDictionary new: rowDictionary size.
	rowDictionary keysAndValuesDo: [:eachTable :eachRowDict |
		| newRowDict |
		newRowDict := IdentityDictionary new: eachRowDict size.
		eachRowDict keysAndValuesDo: [:eachObject :eachRow |
			(self isDeleted: eachRow) ifFalse: [
				newRowDict at: eachObject put: eachRow]].
		newDictionary at: eachTable put: newRowDict].
	rowDictionary := newDictionary.!

deleteRowForTable: aTable withKey: aKey ifAbsent: aBlock
	"Don't actually delete, the fixup of the hash tables is too expensive. Mark invisible instead, overloading the status variable that we know we're not using until later"

	| row |
	row := (self subMapForTable: aTable) at: aKey ifAbsent: aBlock.
	row isNil ifFalse: [self privateDeleteRow: row].!

dictionaryClassRequiredForKey: anObject

	^(self isRowMapKey: anObject)
		ifTrue: [Dictionary]
		ifFalse: [IdentityDictionary].!

differenceFrom: aMementoRowMap
	| differencesMap |
	differencesMap := RowMap new.
	self	additiveDifferencesFrom: aMementoRowMap into: differencesMap.
	self	subtractiveDifferencesFrom: aMementoRowMap into: differencesMap.
	^differencesMap.!

findOrAddRowForTable: aTable withKey: aKey
	| submap row |
	submap := self subMapForTable: aTable withKey: aKey.
	row := submap
		at: aKey
		ifAbsentPut: [DatabaseRow newForTable: aTable withOwner: aKey].
	^(self isDeleted: row)
		ifTrue:
			[| newRow |
			newRow := DatabaseRow newForTable: aTable withOwner: aKey.
			submap at: aKey put: newRow.
			newRow]
		ifFalse: [row].!

includesRowForTable: aTable withKey: aKey

	| row |
	row := (self subMapForTable: aTable ifAbsent: [^false]) at: aKey ifAbsent: [^false].
	^(self isDeleted: row) not.!

initialize
      self initialize: 17!

initialize: size
	rowDictionary := IdentityDictionary new: size.
	hasDeletions := false.!

isDeleted: row
	^row status == #invisible.!

isEmpty
	self compress.
	self rowsDo: [:each | (self isDeleted: each) ifFalse: [^false]].
	^true.!

isRowMapKey: anObject
	^anObject class == RowMapKey or: [anObject class == MultipleRowMapKey].!

isSuitableForCollapsing: eachTablesRows
	"Return true if this set of rows are suitable for collapsing the inserts and deletes together. It's suitable if this is a link table, and if there is a primary key defined which we can use to decide if the rows match or not"
	eachTablesRows isEmpty ifTrue: [^false].
	eachTablesRows
		keysAndValuesDo:
			[:eachKey :eachRow | 
				"(self isRowMapKey: eachKey) ifFalse: [^false]."
				eachRow table primaryKeyFields isEmpty ifTrue: [^false].
				^true ].!

keysAndValuesDo: aBlock 
	self tables do: [:each | 
		(self subMapForTable: each) keysAndValuesDo: aBlock].!

mementoFor: anObject
	^anObject.!

notEmpty
	^self isEmpty not!

numberOfEntries
	self compress.
	^rowDictionary inject: 0 into: [:sum :each | sum + each size]!

numberOfEntriesForTable: aTable
	self compress.
	^(self subMapForTable: aTable) size.!

objects
	| objects |
	objects := IdentitySet new.
	self tables do: [:each | objects addAll: (self subMapForTable: each) keys].
	^objects!

objectsAndRowsDo: aTwoArgumentBlock
	rowDictionary
		do:
			[:eachObjectToRowDictionary | 
			eachObjectToRowDictionary
				keysAndValuesDo:
					[:key :value | 
					(self isDeleted: value)
						ifFalse: [aTwoArgumentBlock value: key value: value]]].!

objectsAndRowsForTable: aTable do: aTwoArgumentBlock
	^(self subMapForTable: aTable)
		keysAndValuesDo:
			[:key :value | 
			(self isDeleted: value)
				ifFalse: [aTwoArgumentBlock value: key value: value]].!

objectsDo: aBlock
	self tables
		do:
			[:each | 
			(self subMapForTable: each)
				keysAndValuesDo:
					[:key :value | (self isDeleted: value) ifFalse: [aBlock value: key]]].!

privateDeleteRow: row
	row status: #invisible.
	hasDeletions := true.!

reverseLookup: anObject

	^anObject.!

rowForTable: aTable withKey: aKey 
	^self rowForTable: aTable withKey: aKey ifAbsent: [self error: 'missing element'].!

rowForTable: aTable withKey: aKey ifAbsent: aBlock
	| row |
	row := (self subMapForTable: aTable) at: aKey ifAbsent: aBlock.
	^(self isDeleted: row) ifTrue: [aBlock value] ifFalse: [row].!

rowForTable: aTable withKey: aKey ifPresent: aBlock
	^(self subMapForTable: aTable)
		at: aKey
		ifPresent:
			[:row | (self isDeleted: row) ifTrue: [nil] ifFalse: [aBlock value: row]].!

rowsDo: aBlock 
	self tables do: [:each | self rowsForTable: each do: aBlock]!

rowsForKey: aKey
	"Return a collection of all rows for any table which are keyed by aKey"

	| rowsForKey |
	rowsForKey := OrderedCollection new: 5.
	rowDictionary
		do:
			[:each | 
			| row |
			row := each at: aKey ifAbsent: [nil].
			(row notNil and: [(self isDeleted: row) not]) ifTrue: [rowsForKey add: row]].
	^rowsForKey.!

rowsForTable: aTable do: aBlock
	^(self subMapForTable: aTable)
		keysAndValuesDo:
			[:key :value | (self isDeleted: value) ifFalse: [aBlock value: value]].!

subMapForTable: aTable
	
	^self subMapForTable: aTable withKey: nil.!

subMapForTable: aTable ifAbsent: aBlock 
	^rowDictionary at: aTable ifAbsent: aBlock!

subMapForTable: aTable withKey: anObject 
	anObject isGlorpProxy ifTrue: [self halt].
	^rowDictionary at: aTable
		ifAbsentPut: [(self dictionaryClassRequiredForKey: anObject) new]!

subtractiveDifferencesFrom: aRowMap into: differencesMap
	"Figure out which things are in aRowMap but not in us. These should be flagged as delete rows. Since we have no further use for aRowMap after this, we can compute this destructively by removing everything that's in us"

	self objectsAndRowsDo: [:object :row |
		aRowMap
			deleteRowForTable: row table
			withKey: object
			ifAbsent: []].
	aRowMap
		objectsAndRowsDo:
			[:object :row | 
			| adjustedObject |
			adjustedObject := (aRowMap reverseLookup: object).
			row forDeletion: true.
			differencesMap
				addRow: row
				forTable: row table
				withKey: adjustedObject].!

tables

	^rowDictionary keys.!

unifyField: eachSourceField key: sourceObject withField: eachTargetField key: targetObject
	| sourceRow targetRow |
	sourceRow := self findOrAddRowForTable: eachSourceField table withKey: sourceObject.
	targetRow := self findOrAddRowForTable: eachTargetField table withKey: targetObject.
	sourceRow unifyField: eachSourceField withField: eachTargetField inRow: targetRow.!

unifyField: sourceField key: sourceObject withField: targetField key: targetObject transformationForward: forwardTransformation inverse: inverseTransformation
	| sourceRow targetRow existingWrapper newWrapper |
	sourceRow := self findOrAddRowForTable: sourceField table withKey: sourceObject.
	targetRow := self findOrAddRowForTable: targetField table withKey: targetObject.
	sourceRow unifyField: sourceField withField: targetField inRow: targetRow.
	existingWrapper := targetRow wrapperAt: targetField.
	(existingWrapper class == TransformingFieldValueWrapper and: [
		existingWrapper forwardTransformation == forwardTransformation and: [
			existingWrapper inverseTransformation == inverseTransformation]])
				ifTrue: [^self].
	newWrapper := (TransformingFieldValueWrapper new)
		innerWrapper: existingWrapper;
		forwardTransformation: forwardTransformation;
		inverseTransformation: inverseTransformation.
	targetRow wrapperAt: targetField put: newWrapper.! !
!RowMap categoriesFor: #additiveDifferencesFrom:into:!private!set operations! !
!RowMap categoriesFor: #addRow:!lookup!private! !
!RowMap categoriesFor: #addRow:forTable:withKey:!lookup!private! !
!RowMap categoriesFor: #adjustForMementos:!private!private/mapping! !
!RowMap categoriesFor: #at:key:put:!api!private! !
!RowMap categoriesFor: #collapseMatchingInsertsAndDeletes!private/deletion!public! !
!RowMap categoriesFor: #collapseMatchingInsertsAndDeletesFor:!private/deletion!public! !
!RowMap categoriesFor: #collectionMementoFor:!private!private/mapping! !
!RowMap categoriesFor: #compress!private!private/deletion! !
!RowMap categoriesFor: #deleteRowForTable:withKey:ifAbsent:!lookup!private! !
!RowMap categoriesFor: #dictionaryClassRequiredForKey:!private!private/mapping! !
!RowMap categoriesFor: #differenceFrom:!private!set operations! !
!RowMap categoriesFor: #findOrAddRowForTable:withKey:!api!private! !
!RowMap categoriesFor: #includesRowForTable:withKey:!lookup!private! !
!RowMap categoriesFor: #initialize!initialize/release!private! !
!RowMap categoriesFor: #initialize:!initialize/release!private! !
!RowMap categoriesFor: #isDeleted:!private!private/deletion! !
!RowMap categoriesFor: #isEmpty!private!testing! !
!RowMap categoriesFor: #isRowMapKey:!api!private! !
!RowMap categoriesFor: #isSuitableForCollapsing:!private!private/deletion! !
!RowMap categoriesFor: #keysAndValuesDo:!iterating!private! !
!RowMap categoriesFor: #mementoFor:!api!private! !
!RowMap categoriesFor: #notEmpty!private!testing! !
!RowMap categoriesFor: #numberOfEntries!counting!private! !
!RowMap categoriesFor: #numberOfEntriesForTable:!counting!private! !
!RowMap categoriesFor: #objects!iterating!private! !
!RowMap categoriesFor: #objectsAndRowsDo:!iterating!private! !
!RowMap categoriesFor: #objectsAndRowsForTable:do:!iterating!private! !
!RowMap categoriesFor: #objectsDo:!iterating!private! !
!RowMap categoriesFor: #privateDeleteRow:!private!private/deletion! !
!RowMap categoriesFor: #reverseLookup:!private!private/mapping! !
!RowMap categoriesFor: #rowForTable:withKey:!lookup!private! !
!RowMap categoriesFor: #rowForTable:withKey:ifAbsent:!lookup!private! !
!RowMap categoriesFor: #rowForTable:withKey:ifPresent:!lookup!private! !
!RowMap categoriesFor: #rowsDo:!iterating!private! !
!RowMap categoriesFor: #rowsForKey:!private!private/mapping! !
!RowMap categoriesFor: #rowsForTable:do:!iterating!private! !
!RowMap categoriesFor: #subMapForTable:!private!private/mapping! !
!RowMap categoriesFor: #subMapForTable:ifAbsent:!private!private/mapping! !
!RowMap categoriesFor: #subMapForTable:withKey:!private!private/mapping! !
!RowMap categoriesFor: #subtractiveDifferencesFrom:into:!private!set operations! !
!RowMap categoriesFor: #tables!private!private/mapping! !
!RowMap categoriesFor: #unifyField:key:withField:key:!api!private! !
!RowMap categoriesFor: #unifyField:key:withField:key:transformationForward:inverse:!api!private! !

!RowMap class methodsFor!

forRows: aCollection
	"Convenience - construct a row map from a given set of rows"
	| map |
	map := self new.
	aCollection do: [:each | map addRow: each].
	^map.!

new

	^super new initialize.!

new: size

	^super new initialize: size.! !
!RowMap class categoriesFor: #forRows:!instance creation!private! !
!RowMap class categoriesFor: #new!instance creation!private! !
!RowMap class categoriesFor: #new:!instance creation!private! !

RowMapKey guid: (GUID fromString: '{CADF5555-3CA8-460A-8836-D233B7B0D8AA}')!
RowMapKey comment: 'This class serves as a key for a dictionary containing two sub-keys, where we want to be able to look up based on the identity of both sub-keys paired together. This is used primarily for many-to-many mappings indexing into rowmaps, where we want to key the row by the identity of the object that determines it, but there are two of them.

Instance Variables:

key1	<Object>	One sub-key.
key2	<Object>	The other sub-key.'''!
!RowMapKey categoriesForClass!Unclassified! !
!RowMapKey methodsFor!

= aRowMapKey

	aRowMapKey class == self class ifFalse: [^false].
	^(key1 == aRowMapKey key1 and: [key2 == aRowMapKey key2]) or: [
		key2 == aRowMapKey key1 and: [key1 == aRowMapKey key2]].!

adjustIn: aCorrespondenceMap
	"Adjust ourselves to refer to the copies rather than the originals"
	| adjusted |
	adjusted := self copy.
	adjusted key1: (aCorrespondenceMap at: key1 ifAbsent: [key1]).
	adjusted key2: (aCorrespondenceMap at: key2 ifAbsent: [key2]).
	^adjusted.!

hash
	| hash1 hash2 hash1Temp longHash |
	hash1 := key1 identityHash.
	hash2 := key2 identityHash.
	hash1 > hash2 ifTrue: [
		hash1Temp := hash1.
		hash1 := hash2.
		hash2 := hash1Temp].
	longHash := (hash1 bitShift: 8) bitXor: hash2.
	^Dialect fitHashIntoSmallInteger: longHash.!

key1
	^key1!

key1: anObject
	key1 := anObject!

key2
	^key2!

key2: anObject
	key2 := anObject!

reverseAdjustIn: aRowMapForMementos
	"Adjust ourselves to refer to the copies rather than the originals"
	| adjusted |
	adjusted := self copy.
	adjusted key1: (aRowMapForMementos reversedCorrespondenceMap at: key1 ifAbsent: [key1]).
	adjusted key2: (aRowMapForMementos reversedCorrespondenceMap at: key2 ifAbsent: [key2]).
	^adjusted.! !
!RowMapKey categoriesFor: #=!comparing!private! !
!RowMapKey categoriesFor: #adjustIn:!adjusting!private! !
!RowMapKey categoriesFor: #hash!comparing!private! !
!RowMapKey categoriesFor: #key1!accessing!private! !
!RowMapKey categoriesFor: #key1:!accessing!private! !
!RowMapKey categoriesFor: #key2!accessing!private! !
!RowMapKey categoriesFor: #key2:!accessing!private! !
!RowMapKey categoriesFor: #reverseAdjustIn:!adjusting!private! !

TableSorter guid: (GUID fromString: '{353E6594-A121-491B-AF65-6BBE32647FFA}')!
TableSorter comment: 'This implements a topological sort for tables, based on the foreign key constraints between them.

Instance Variables:
	orderedTables	<Collection of: DatabaseTable>	The resulting tables in sorted order.
	tables	<(Collection of: (DatabaseTable))>	The input tables.
	visitedTables	<IdentitySet of: DatabaseTable>	The tables we have already visited during the sort.

'!
!TableSorter categoriesForClass!Unclassified! !
!TableSorter methodsFor!

addTable: aTable 
	tables add: aTable!

hasBeenVisited: aTable

	^visitedTables includes: aTable.!

initialize
	tables := OrderedCollection new: 100.
	visitedTables := IdentitySet new: 100.!

markVisited: aTable

	visitedTables add: aTable.!

sort
	orderedTables := OrderedCollection new: tables size.
	tables do: [:each | self visit: each].
	^orderedTables select: [:each | tables includes: each].!

visit: aTable 
	"The essential bit of topological sort. Visit each node in post-order, traversing dependencies, based on foreign key constraints to database-generated fields. "

	(self hasBeenVisited: aTable) ifTrue: [^self].
	self markVisited: aTable.
	self visitDependentTablesFor: aTable.
	orderedTables add: aTable!

visitDependentTablesFor: aTable 
	aTable foreignKeyConstraints do: [:eachConstraint | 
          self visit: eachConstraint targetTable.]! !
!TableSorter categoriesFor: #addTable:!accessing!private! !
!TableSorter categoriesFor: #hasBeenVisited:!accessing!private! !
!TableSorter categoriesFor: #initialize!initializing!private! !
!TableSorter categoriesFor: #markVisited:!accessing!private! !
!TableSorter categoriesFor: #sort!private!sorting! !
!TableSorter categoriesFor: #visit:!private!sorting! !
!TableSorter categoriesFor: #visitDependentTablesFor:!private!sorting! !

!TableSorter class methodsFor!

for: tables

	| sorter |
	sorter := self new.
	tables do: [:each |
		sorter addTable: each].
	^sorter.!

new

	^super new initialize.! !
!TableSorter class categoriesFor: #for:!instance creation!private! !
!TableSorter class categoriesFor: #new!instance creation!private! !

UnitOfWork guid: (GUID fromString: '{2D2AA39F-37BB-4F49-AF76-AF8E9C36349C}')!
UnitOfWork comment: 'A UnitOfWork keeps track of objects which might potentially be modified and lets you roll them back or commit the changes into the database.

Instance Variables:
	newObjects	<IdentitySet of: Object>	The objects registered with us as being new (i.e. not existing already in the database)
	session	<Session>	The session in which this is all taking place.
	transaction	<ObjectTransaction>	Keeps track of the original object state so that we can revert it.
	rowMap	<RowMap>	A holder for the rows when we are writing out changes.
	commitPlan	<(OrderedCollection of: DatabaseRow)>	The list of rows to be written, in order. Constructed by topological sorting the contents of the row map.
	calculatingRows <Boolean> Are we in the process of computing the rows to be written. If so, we should not be doing reads and registrations - it''s probably an error in the mapping code by accidentally forcing a read.

'!
!UnitOfWork categoriesForClass!Unclassified! !
!UnitOfWork methodsFor!

abort
	self reinitialize.!

abort: anObject
	transaction abort: anObject!

addObject: eachObject toCacheKeyedBy: key 

	| tracker |	
	self session cacheAt: key put: eachObject.
	tracker := self session modificationTracker.
	tracker isNil ifFalse: [tracker track: eachObject]!

addToCommitPlan: aRow
	aRow shouldBeWritten ifFalse: [^self].
	commitPlan add: aRow.!

addToDeletePlan: aRow
	
	deletePlan add: aRow.!

assignSequenceNumbers
	commitPlan do: [:each | each preWriteAssignSequencesUsing: self session].!

begin
	self reinitialize.!

buildCommitPlan
	| tablesInCommitOrder |
	commitPlan := OrderedCollection new.
	deletePlan := OrderedCollection new.
	tablesInCommitOrder := session tablesInCommitOrder.
	tablesInCommitOrder do: 
			[:eachTable | 
			self rowsForTable: eachTable
				do: 
					[:eachRow | 
					eachRow forDeletion 
						ifTrue: [self addToDeletePlan: eachRow]
						ifFalse: [self addToCommitPlan: eachRow]]].
	self sortCommitPlanRows.!

commit
	[self preCommit.
	self writeRows] ifCurtailed: [succeeded := false].
	succeeded
		ifTrue: [self postCommit]
		ifFalse:
			[GlorpTransactionFailure signal].!

computeCollectionDeletes
	"If we have objects that are part of exclusive collections, and they have been removed from those collections, then they must be deleted. Note that we do this *before* we register the transitive closure at the end of the unit of work. That means we won't know about any new objects yet, but we don't care because new objects won't be deleted in this way. And that way if we delete and then add another object with (ick) the same primary key, we shouldn't have a clash because the delete will already have been issued."

	self registeredObjects do: [:each |
		| descriptor |
		(descriptor := session descriptorFor: each class) isNil
			ifFalse: [descriptor mappings do: [:eachMapping |
				self computeCollectionDeletesFor: eachMapping inObject: each]]].!

computeCollectionDeletesFor: aMapping inObject: anObject
	"Figure out which, if any, dependent objects we should delete based on changes to the corresponding collection in anObject"
	| collection memento collectionMemento |
	aMapping isExclusive ifFalse: [^self].
	aMapping attribute isCollectionAttribute ifFalse: [^self].
	collection := aMapping getValueFrom: anObject.
	collection := self session realObjectFor: collection ifNone: [^self].
	(aMapping isValidTarget: collection) ifFalse: [^self].
	memento := transaction mementoFor: anObject.
	collectionMemento := transaction collectionMementoFor: (self session realObjectFor: (aMapping getValueFrom: memento)).
	"OK, now we have the state of the original collection and the new collection, we can look for differences."
	collectionMemento do: [:each |
		| matching |
		(each isGlorpProxy not or: [each isInstantiated]) ifTrue: [
			matching := collection 
				detect: [:eachElement | 
					(eachElement isGlorpProxy not or: [eachElement isInstantiated])
						and: [each yourSelf == eachElement yourSelf]]
				ifNone: [nil].
			matching == nil ifTrue: [session delete: each]
		]
	].!

correspondenceMap
	^transaction undoMap!

createMementoRowMapFor: objects 
	"Create a rowmap for the objects whose state was already known. We subtract this from the rowmap of all known objects to get the rows that need to be written. New objects are also registered, so we only generate rows here for non-new objects"

	| localRowMap |
	localRowMap := RowMapForMementos 
				withCorrespondenceMap: self correspondenceMap.
	localRowMap collectionMementos: transaction collectionMementos.
	objects keysAndValuesDo: 
			[:original :memento | 
			(self newObjects includes: original) 
				ifFalse: [session createRowsFor: memento in: localRowMap]].
	^localRowMap!

createRowMapFor: objects
	| localRowMap |
	localRowMap := RowMap new: objects size.
	objects do: [:each | session createRowsFor: each in: localRowMap].
	^localRowMap!

createRows
	self createRowsForPartialWrites.!

createRowsForCompleteWrites
	"reference implementation. not called from anywhere"

	self registeredObjectsDo: [:eachObject | session createRowsFor: eachObject in: rowMap].
	deletedObjects do: [:eachObject | session createDeleteRowsFor: eachObject in: rowMap].!

createRowsForPartialWrites
	| registeredObjectsRowMap mementoObjectsRowMap |
	registeredObjectsRowMap := self createRowMapFor: self registeredObjects.
	deletedObjects 
		do: [:eachObject | session createDeleteRowsFor: eachObject in: registeredObjectsRowMap].
	mementoObjectsRowMap := self createMementoRowMapFor: self mementoObjects.
	rowMap := registeredObjectsRowMap differenceFrom: mementoObjectsRowMap.!

delete: anObject
	(self isRegistered: anObject) ifFalse: [session registerForDelete: anObject].
	deletedObjects add: anObject.
	(session embeddedObjectsFor: anObject) do: [:each | self delete: each].!

deletedObjects
	^deletedObjects.!

forcesUpdateFor: aDatabaseRow
	^linkRowsForUpdate notNil and: [linkRowsForUpdate includes: aDatabaseRow].!

hasPendingDeletions

	^deletedObjects isEmpty not.!

initialize

	transaction := ObjectTransaction new.
	self reinitialize!

instantiateIfRequiredFor: anObject mapping: eachMapping
	"Sometimes we have to instantiate the targets if they weren't. Specifically, if there's a relationship where the target has a foreign key to us. e.g. if X has a 1-many to Y, and we don't instantiate the collection of Y, but then replace it with some other collection. The Y's keys have to be updated, so we need to make sure they're read. There are other cases, too"

	| originalTarget targetObject mapping instantiated |
	mapping := eachMapping applicableMappingForObject: anObject.
	mapping isRelationship ifFalse: [^false].
	mapping readOnly ifTrue: [^false].
	targetObject := mapping getValueFrom: anObject.
	(self isNewObject: anObject)
		ifTrue:
			[^self instantiateIfRequiredForNewObjectReferenceTo: targetObject].
	instantiated := (self willDelete: anObject)
		ifTrue:
			[self instantiateIfRequiredForDeletedObjectReferencing: targetObject from: anObject in: eachMapping]
		ifFalse: [false].
	originalTarget := self originalValueFor: anObject mapping: mapping.
	^instantiated or: [self
		instantiateIfRequiredForReferenceChangedFrom: originalTarget
		to: targetObject].!

instantiateIfRequiredForDeletedObjectReferencing: targetObject from: anObject in: aMapping
	"If we are deleted, then we need to have at least one mapping for each of our primary keys instantiated. If we're lucky, there's a direct mapping, and it's easy. Note that this only looks for the primary table. This might not be correct, but it's not clear, and for secondary tables there's a multiple table join that can be used without needing a mapping. And I'm not prepared to think about secondary tables whose keys are mapped only as relationships. In addition, if we use a link table, then we need to instantiate the reference, because we'll need to have enough information to delete the link table entry/entries."
	| mappingsToUseForPrimaryKeys |
	aMapping mapsPrimaryKeys ifFalse: [^false].	
	aMapping usesLinkTable ifTrue: [
		"Make sure we get the non-filtered version, and return false so that if the reference has changed, we'll also get the original version."
		self notFilteringDeletesDo: [targetObject yourSelf]. ^false].
	mappingsToUseForPrimaryKeys := aMapping descriptor primaryKeyMappingsFor: aMapping descriptor primaryTable.
	"For any of the fields we map, is there anyone else who can get us the value. If not, instantiate us. It might be better to optimize this a bit to look for e.g. one-one mappings before to-many."
	aMapping mappedFields do: [:each |
		| mappings |
		mappings := mappingsToUseForPrimaryKeys at: each ifAbsent: [nil].
		mappings isNil ifFalse: [
			(mappings allSatisfy: [:eachMapping | 
				(eachMapping hasValueThatCanProvideField: each from: anObject) not])
					ifTrue: [
						targetObject yourSelf.
						^true]]].
	^false.!

instantiateIfRequiredForNewObjectReferenceTo: targetObject
	"If it's a new object, then we always need to instantiate"
	targetObject isGlorpProxy ifFalse: [^false].
	targetObject getValue.
	^true.!

instantiateIfRequiredForReferenceChangedFrom: originalTarget to: newTarget
	"For an existing relationship, we need to instantiate if the object we're referring to changed, and either was or is now a proxy. Actually we might get away without if it used to be, but now isn't, but better safe than sorry."

	| wasProxy  |
	originalTarget == newTarget ifTrue: [^false].
	wasProxy := false.
	self notFilteringDeletesDo: [
		originalTarget isGlorpProxy ifTrue: [wasProxy := true. originalTarget getValue].
		newTarget isGlorpProxy ifTrue: [wasProxy := true. newTarget getValue]].
	^wasProxy.!

isNewObject: anObject 
	^self newObjects includes: anObject.!

isRegistered: anObject

	^transaction isRegistered: anObject.!

mementoObjects
	"Warning: Excessive cleverness!!!!!! The mementoObjects we want to iterate over are the values in the correspondenceMap dictionary. We were getting the values and returning them, but if all we need to do is iterate, then the dictionary itself works fine"
	^self correspondenceMap.!

newObjects
	newObjects isNil ifTrue: [newObjects := IdentitySet new].
	^newObjects!

notFilteringDeletesDo: aBlock
	"Execute aBlock, with filtering of deletes turned off. This is useful if we are in the process of deleting an object. Suppose we are deleting an object that has a link table relationship, but we have already deleted one of the linked objects. We need to make sure we also delete the link table entry, which means we need to know about that linked object, so we don't want Glorp filtering it out for us."

	| oldCommitPhase |
	[oldCommitPhase := commitPhase.
	commitPhase := #registering.
	aBlock value] ensure: [commitPhase := oldCommitPhase].!

numberOfRows
	^commitPlan size + deletePlan size.!

originalValueFor: anObject mapping: eachMapping
	| memento |
	memento := transaction undoMap at: anObject.
	^eachMapping getValueFrom: memento.!

postCommit
	self sendPostWriteNotification.
	self updateSessionCache.
	self validateRowsPostWrite.!

preCommit
	commitPhase := #registering.
	self registerTransitiveClosure.
	commitPhase := #building.
	self createRows.
	linkRowsForUpdate := rowMap collapseMatchingInsertsAndDeletes.
	self buildCommitPlan.
	commitPhase := nil.
	self sendPreWriteNotification.
	self validateRows.!

privateGetRowMap

	^rowMap!

privateGetTransaction

	^transaction.!

propagateDeletes
	"If we have objects which are to be deleted, and they have exclusive relationships to other objects, then we must also delete those other objects."
	| deletionQueue |
	deletionQueue := OrderedCollection new: deletedObjects size.
	deletionQueue addAll: deletedObjects.
	[deletionQueue isEmpty]
		whileFalse:
			[| each |
			each := deletionQueue removeFirst.
			(self session descriptorFor: each) mappings
				do:
					[:eachMapping | 
					eachMapping isExclusive
						ifTrue:
							[eachMapping
								loopOverAttributeOf: each
								doing:
									[:index :subObject | 
									| realSubObject |
									realSubObject := subObject yourSelf.
									(realSubObject notNil and: [(deletedObjects includes: realSubObject) not])
										ifTrue:
											[session delete: subObject.
											deletionQueue add: realSubObject]]]]].!

readBackNewRowInformation
	| changedObjects |
	changedObjects := rowMap objects.
	changedObjects
		do: 
			[:each | 
			| descriptor |
			descriptor := session descriptorFor: each.
			descriptor isNil ifFalse: [descriptor readBackNewRowInformationFor: each in: rowMap]]!

register: anObject 

	| realObject |
	commitPhase == #building ifTrue: [self halt]. "Should not happen. Probably indicates that we're triggering proxies while firing the mappings"
	realObject := transaction register: anObject.
	self registerTransitiveClosureFrom: realObject.!

registerAsNew: anObject 
	anObject isNil ifTrue: [^nil].
	commitPhase == #building ifTrue: [self halt]. "Should not happen. Probably indicates that we're triggering proxies while firing the mappings"
	self newObjects add: anObject.
	self register: anObject.
	^anObject!

registeredObjects
	^self correspondenceMap keys!

registeredObjectsDo: aBlock

	transaction registeredObjectsDo: [:each |
		(each glorpIsCollection or: [session hasDescriptorFor: each]) ifTrue: [
			aBlock value: each]].!

registerForDelete: anObject 
	"Register anObject with intent to delete it."
	[deleting add: anObject.
	self register: anObject] ensure: [deleting remove: anObject ifAbsent: []].!

registerTransitiveClosure
	"Look for new objects reachable from currently registered objects, and propagate deletions"

	self computeCollectionDeletes.
	self
		registeredObjectsDo:
			[:eachObject | self registerTransitiveClosureFrom: eachObject].
	self propagateDeletes.!

registerTransitiveClosureFrom: anObject

	| descriptor |
	anObject glorpIsCollection ifTrue: [
		anObject do: [:each | session register: each].
		^self].

	descriptor := session descriptorFor: anObject class.
	descriptor isNil ifTrue: [^self].
	descriptor mappings do: [:eachMapping |
			self instantiateIfRequiredFor: anObject mapping: eachMapping].
	descriptor referencedIndependentObjectsFrom: anObject do: [:eachObject |
		session register: eachObject].!

reinitialize

	rowMap := RowMap new.
	commitPhase := nil.
	deletedObjects := IdentitySet new.
	deleting := IdentitySet new.
	commitPlan := OrderedCollection new.
	deletePlan := OrderedCollection new.
	transaction abort.!

relatedRowsFor: each
	| rows extraRows |
	rows := each relatedRowsIn: rowMap.
	extraRows := self session system additionalRelatedRowsFor: each in: rowMap.
	extraRows isNil ifFalse: [rows addAll: extraRows].
	^rows select: [:eachRelatedRow | eachRelatedRow status notNil].!

reserveSequenceNumbers
	rowsByTable := commitPlan glorpGroupedBy: [:each | each table].
	rowsByTable keysAndValuesDo: [:eachTable :rows | 
		eachTable sequences do: [:eachSequence | 
			eachSequence 
				reserveSequenceNumbers: (
					rows inject: 0 into: [:sum :each | 
						(each needsValueFromSequence: eachSequence) 
							ifTrue: [sum + 1] ifFalse: [sum]])
					in: self session
					for: eachTable]]!

rollback
	self abort.!

rowsForTable: aTable do: aBlock

	rowMap rowsForTable: aTable do: aBlock.!

rowsToWrite
	^commitPlan.!

sendPostWriteNotification
	self 
		registeredObjectsDo: [:eachObject | session sendPostWriteEventTo: eachObject]!

sendPreWriteNotification
	self 
		registeredObjectsDo: [:eachObject | session sendPreWriteEventTo: eachObject]!

session
	^session!

session: aGlorpSession
	session := aGlorpSession!

shouldFilterDeletions
	"Should we filter deleted objects from the results of queries. We don't do this if we're in the process of registering the transitive closure, because any fetches we do there are in order to get information we need for the commit (e.g. primary keys). If we filtered the deleted objects, then we may not be able to figure out primary keys for the objects to be deleted"
	^commitPhase ~~ #registering.!

sortCommitPlanRows
	| sorter |
	sorter := GlorpRowSorter new
		getChildrenVia: [:each | self relatedRowsFor: each].
	commitPlan := sorter sort: commitPlan.
	sorter := GlorpRowSorter new
		getChildrenVia: [:each | self relatedRowsFor: each].
	deletePlan := sorter sort: deletePlan.!

system
	^session system.!

undelete: anObject ifNotDeleted: aBlock
	deletedObjects remove: anObject ifAbsent: aBlock.
	(session embeddedObjectsFor: anObject) do: [:each | self undelete: each].!

updateSessionCache
	rowMap
		keysAndValuesDo: [:eachObject :eachRow | 
			eachRow shouldBeWritten ifTrue: [
				self updateSessionCacheFor: eachObject withRow: eachRow]].
	deletedObjects do: [:each |
		session cacheRemoveObject: each].
	linkRowsForUpdate do: [:eachRow |
		eachRow shouldBeWritten ifTrue: [
			self updateSessionCacheFor: eachRow owner withRow: eachRow]].!

updateSessionCacheFor: anObject withRow: aRow 
	| key |
	(rowMap isRowMapKey: anObject) ifTrue: [^self].  "Not cachable"
	key := aRow primaryKey.
	key isNil ifTrue: [^self].
	(session cacheContainsObjectForClass: anObject class key: key) 
		ifFalse: [self addObject: anObject toCacheKeyedBy: key].!

validateRows
	"Perform basic validation. Always test for equal named but non-identical tables, a sign of a malformed  system or other loss of identity. Also delegate to the descriptorSystem, which may have useful application-specific validations to do."
	
	| tables tableNames |
	tables := Set new.
	rowMap rowsDo: [:each |
		tables add: each table].
	tableNames := tables collect: [:each | each qualifiedName].
	tables asSet size = tableNames asSet size ifFalse: [self error: 'multiple table objects with the same name'].
	self system validateRows: self.!

validateRowsPostWrite
	"Perform basic validation after writing is done. It's better to validate beforehand, but some things are easier to catch afterwards"
	
	self system validateRowsPostWrite: self.!

willDelete: anObject
	"Return true if we will delete anObject when we commit."	
	^(deletedObjects includes: anObject) or: [deleting includes: anObject].!

writeRows

	succeeded := true.
	self reserveSequenceNumbers.
	self assignSequenceNumbers.
	self session tablesInCommitOrder do: [:eachTable | self writeRowsForTable: eachTable].
	deletePlan reverseDo: [:eachRow | | rowSucceeded |
		rowSucceeded := session writeRow: eachRow.
		succeeded := succeeded & rowSucceeded].
	self readBackNewRowInformation.!

writeRowsForTable: aTable

	| rows rowsByOperation |

	rows := rowsByTable at: aTable ifAbsent: [#()].
	^rows isEmpty ifFalse: [
		rowsByOperation := rows glorpGroupedBy: [:eachRow | session shouldInsert: eachRow].
		rowsByOperation do: [:eachListOfHomogeneousRows | 
			succeeded := succeeded & (session writeHomogeneousRows: eachListOfHomogeneousRows)]].! !
!UnitOfWork categoriesFor: #abort!begin/commit/abort!private! !
!UnitOfWork categoriesFor: #abort:!begin/commit/abort!private! !
!UnitOfWork categoriesFor: #addObject:toCacheKeyedBy:!private!private/mapping! !
!UnitOfWork categoriesFor: #addToCommitPlan:!private!private/mapping! !
!UnitOfWork categoriesFor: #addToDeletePlan:!private!private/mapping! !
!UnitOfWork categoriesFor: #assignSequenceNumbers!private!private/mapping! !
!UnitOfWork categoriesFor: #begin!begin/commit/abort!private! !
!UnitOfWork categoriesFor: #buildCommitPlan!private!private/mapping! !
!UnitOfWork categoriesFor: #commit!begin/commit/abort!private! !
!UnitOfWork categoriesFor: #computeCollectionDeletes!deletion!private! !
!UnitOfWork categoriesFor: #computeCollectionDeletesFor:inObject:!deletion!private! !
!UnitOfWork categoriesFor: #correspondenceMap!accessing!private! !
!UnitOfWork categoriesFor: #createMementoRowMapFor:!begin/commit/abort!private! !
!UnitOfWork categoriesFor: #createRowMapFor:!begin/commit/abort!private! !
!UnitOfWork categoriesFor: #createRows!begin/commit/abort!private! !
!UnitOfWork categoriesFor: #createRowsForCompleteWrites!begin/commit/abort!private! !
!UnitOfWork categoriesFor: #createRowsForPartialWrites!begin/commit/abort!private! !
!UnitOfWork categoriesFor: #delete:!deletion!private! !
!UnitOfWork categoriesFor: #deletedObjects!private! !
!UnitOfWork categoriesFor: #forcesUpdateFor:!private!testing! !
!UnitOfWork categoriesFor: #hasPendingDeletions!deletion!private! !
!UnitOfWork categoriesFor: #initialize!initializing!private! !
!UnitOfWork categoriesFor: #instantiateIfRequiredFor:mapping:!private!private/mapping! !
!UnitOfWork categoriesFor: #instantiateIfRequiredForDeletedObjectReferencing:from:in:!private!private/mapping! !
!UnitOfWork categoriesFor: #instantiateIfRequiredForNewObjectReferenceTo:!private!private/mapping! !
!UnitOfWork categoriesFor: #instantiateIfRequiredForReferenceChangedFrom:to:!private!private/mapping! !
!UnitOfWork categoriesFor: #isNewObject:!private!testing! !
!UnitOfWork categoriesFor: #isRegistered:!private!registering! !
!UnitOfWork categoriesFor: #mementoObjects!begin/commit/abort!private! !
!UnitOfWork categoriesFor: #newObjects!accessing!private! !
!UnitOfWork categoriesFor: #notFilteringDeletesDo:!private! !
!UnitOfWork categoriesFor: #numberOfRows!accessing!private! !
!UnitOfWork categoriesFor: #originalValueFor:mapping:!private!private/mapping! !
!UnitOfWork categoriesFor: #postCommit!begin/commit/abort!private! !
!UnitOfWork categoriesFor: #preCommit!begin/commit/abort!private! !
!UnitOfWork categoriesFor: #privateGetRowMap!private! !
!UnitOfWork categoriesFor: #privateGetTransaction!private! !
!UnitOfWork categoriesFor: #propagateDeletes!begin/commit/abort!private! !
!UnitOfWork categoriesFor: #readBackNewRowInformation!private!private/mapping! !
!UnitOfWork categoriesFor: #register:!private!registering! !
!UnitOfWork categoriesFor: #registerAsNew:!private! !
!UnitOfWork categoriesFor: #registeredObjects!begin/commit/abort!private! !
!UnitOfWork categoriesFor: #registeredObjectsDo:!enumerating!private! !
!UnitOfWork categoriesFor: #registerForDelete:!private!registering! !
!UnitOfWork categoriesFor: #registerTransitiveClosure!begin/commit/abort!private! !
!UnitOfWork categoriesFor: #registerTransitiveClosureFrom:!private!private/mapping! !
!UnitOfWork categoriesFor: #reinitialize!initializing!private! !
!UnitOfWork categoriesFor: #relatedRowsFor:!private!private/mapping! !
!UnitOfWork categoriesFor: #reserveSequenceNumbers!private!private/mapping! !
!UnitOfWork categoriesFor: #rollback!begin/commit/abort!private! !
!UnitOfWork categoriesFor: #rowsForTable:do:!enumerating!private! !
!UnitOfWork categoriesFor: #rowsToWrite!accessing!private! !
!UnitOfWork categoriesFor: #sendPostWriteNotification!private! !
!UnitOfWork categoriesFor: #sendPreWriteNotification!private! !
!UnitOfWork categoriesFor: #session!accessing!private! !
!UnitOfWork categoriesFor: #session:!accessing!private! !
!UnitOfWork categoriesFor: #shouldFilterDeletions!private!testing! !
!UnitOfWork categoriesFor: #sortCommitPlanRows!private!private/mapping! !
!UnitOfWork categoriesFor: #system!accessing!private! !
!UnitOfWork categoriesFor: #undelete:ifNotDeleted:!deletion!private! !
!UnitOfWork categoriesFor: #updateSessionCache!private!private/mapping! !
!UnitOfWork categoriesFor: #updateSessionCacheFor:withRow:!private!private/mapping! !
!UnitOfWork categoriesFor: #validateRows!begin/commit/abort!private! !
!UnitOfWork categoriesFor: #validateRowsPostWrite!begin/commit/abort!private! !
!UnitOfWork categoriesFor: #willDelete:!deletion!private! !
!UnitOfWork categoriesFor: #writeRows!private!private/mapping! !
!UnitOfWork categoriesFor: #writeRowsForTable:!private!private/mapping! !

!UnitOfWork class methodsFor!

new

	^super new initialize.! !
!UnitOfWork class categoriesFor: #new!instance creation!private! !

TransformingFieldValueWrapper guid: (GUID fromString: '{F5D7E2CE-B0DC-4ABB-83BB-F2BF50A0B5E0}')!
TransformingFieldValueWrapper comment: 'This wraps another wrapper and transforms its value.

Instance Variables:
	forwardTransformation	<Block>	What to do to convert our value going out.
	innerWrapper	<FieldValueWrapper>	The wrapper that really has the value.
	inverseTransformation	<Block>	What to do to convert our value going in.

'!
!TransformingFieldValueWrapper categoriesForClass!Unclassified! !
!TransformingFieldValueWrapper methodsFor!

allInnerWrappers
	^innerWrapper allInnerWrappers, (Array with: innerWrapper).!

basicContents
	"Return the un-transformed contents"
	^innerWrapper contents.!

containsWrapper: aFieldValueWrapper
	^self == aFieldValueWrapper or: [innerWrapper == aFieldValueWrapper
		or: [aFieldValueWrapper innerWrapper == self
			or: [aFieldValueWrapper innerWrapper == self innerWrapper]]].!

contents

	^forwardTransformation value: innerWrapper contents.!

contents: anObject

	innerWrapper contents: (inverseTransformation value: anObject).!

forwardTransformation
	^forwardTransformation!

forwardTransformation: aBlock
	forwardTransformation := aBlock!

hasValue

	^innerWrapper hasValue.!

innerWrapper
	^innerWrapper!

innerWrapper: aFieldValueWrapper
	innerWrapper := aFieldValueWrapper!

inverseTransformation
	^inverseTransformation!

inverseTransformation: aBlock
	inverseTransformation := aBlock!

privateSetContents: anObject
	innerWrapper privateSetContents: (inverseTransformation value: anObject).!

replaceWith: aFieldValueWrapper
	innerWrapper replaceWith: aFieldValueWrapper.! !
!TransformingFieldValueWrapper categoriesFor: #allInnerWrappers!accessing!private! !
!TransformingFieldValueWrapper categoriesFor: #basicContents!accessing!private! !
!TransformingFieldValueWrapper categoriesFor: #containsWrapper:!private!unifying! !
!TransformingFieldValueWrapper categoriesFor: #contents!accessing!private! !
!TransformingFieldValueWrapper categoriesFor: #contents:!accessing!private! !
!TransformingFieldValueWrapper categoriesFor: #forwardTransformation!accessing!private! !
!TransformingFieldValueWrapper categoriesFor: #forwardTransformation:!accessing!private! !
!TransformingFieldValueWrapper categoriesFor: #hasValue!accessing!private! !
!TransformingFieldValueWrapper categoriesFor: #innerWrapper!accessing!private! !
!TransformingFieldValueWrapper categoriesFor: #innerWrapper:!accessing!private! !
!TransformingFieldValueWrapper categoriesFor: #inverseTransformation!accessing!private! !
!TransformingFieldValueWrapper categoriesFor: #inverseTransformation:!accessing!private! !
!TransformingFieldValueWrapper categoriesFor: #privateSetContents:!accessing!private! !
!TransformingFieldValueWrapper categoriesFor: #replaceWith:!private!unifying! !

RowMapForMementos guid: (GUID fromString: '{BDEAC567-FE78-4FC7-94EC-18C95D847587}')!
RowMapForMementos comment: 'This is a specialized version of RowMap for creating rowmaps out of the mementos in the undo/correspondence map. When doing partial writes we create a rowmap for the current state of the objects, then a rowmap for the original state, and difference the two.

The tricky part is that the mementos refer back to the original objects, so when we establish unification constraints between rows, they would establish them to original objects. This is wrong, and not trivial to debug.

So this rowmap keeps the correspondence map and knows that it has to compensate and get the memento for any related objects.
'!
!RowMapForMementos categoriesForClass!Unclassified! !
!RowMapForMementos methodsFor!

adjustForMementos: objects
	"We may be being passed row map keys that refer to originals when they should refer to mementos. Fix.This is horribly ugly."

	1 to: objects size do: [:i | | possibleRowMapKey |
		possibleRowMapKey := objects at: i.
		(self isRowMapKey: possibleRowMapKey) ifTrue: [
			objects at: i put: (possibleRowMapKey adjustIn: correspondenceMap)]].
	^objects.!

collectionMementoFor: aCollection

	aCollection glorpIsCollection ifFalse: [^aCollection].
	^collectionMementos at: aCollection ifAbsent: [correspondenceMap at: aCollection].!

collectionMementos: aDictionary
	collectionMementos := aDictionary.!

correspondenceMap
	^correspondenceMap!

correspondenceMap: anObject
	correspondenceMap := anObject!

deleteRowForTable: aTable withKey: aKey ifAbsent: aBlock
	^super deleteRowForTable: aTable withKey: (self mementoFor: aKey) ifAbsent: aBlock.!

findOrAddRowForTable: aTable withKey: aKey 
	^super findOrAddRowForTable: aTable withKey: (self mementoFor: aKey).!

mementoFor: aKey
	^(self isRowMapKey: aKey)
		ifTrue: [aKey adjustIn: correspondenceMap]
		ifFalse: [correspondenceMap at: aKey ifAbsent: [aKey]].!

originalObjectFor: anObject

	^correspondenceMap at: anObject.!

reversedCorrespondenceMap
	(reversedCorrespondenceMap isNil
		or: [reversedCorrespondenceMap size ~= correspondenceMap size])
		ifTrue:
			[reversedCorrespondenceMap := IdentityDictionary
				new: correspondenceMap size.
			correspondenceMap
				keysAndValuesDo: [:k :v | reversedCorrespondenceMap at: v put: k]].
	^reversedCorrespondenceMap.!

reverseLookup: anObject

      (self isRowMapKey: anObject) ifTrue: [
            ^anObject reverseAdjustIn: self].
      ^ self reversedCorrespondenceMap at: anObject.!

rowForTable: aTable withKey: aKey ifAbsent: aBlock
	^super
		rowForTable: aTable
		withKey: (self mementoFor: aKey)
		ifAbsent: aBlock.! !
!RowMapForMementos categoriesFor: #adjustForMementos:!api!private! !
!RowMapForMementos categoriesFor: #collectionMementoFor:!private!private/mapping! !
!RowMapForMementos categoriesFor: #collectionMementos:!accessing!private! !
!RowMapForMementos categoriesFor: #correspondenceMap!accessing!private! !
!RowMapForMementos categoriesFor: #correspondenceMap:!accessing!private! !
!RowMapForMementos categoriesFor: #deleteRowForTable:withKey:ifAbsent:!api!private! !
!RowMapForMementos categoriesFor: #findOrAddRowForTable:withKey:!api!private! !
!RowMapForMementos categoriesFor: #mementoFor:!api!private! !
!RowMapForMementos categoriesFor: #originalObjectFor:!private!private/mapping! !
!RowMapForMementos categoriesFor: #reversedCorrespondenceMap!api!private! !
!RowMapForMementos categoriesFor: #reverseLookup:!private!private/mapping! !
!RowMapForMementos categoriesFor: #rowForTable:withKey:ifAbsent:!api!private! !

!RowMapForMementos class methodsFor!

withCorrespondenceMap: aDictionary

	^self new
		correspondenceMap: aDictionary.! !
!RowMapForMementos class categoriesFor: #withCorrespondenceMap:!instance creation!private! !

"Binary Globals"!

